from bpy.props import (EnumProperty, PointerProperty)
from bpy.types import (Panel, PropertyGroup)
from bpy.app.handlers import persistent
import mathutils
import numpy as np
import math
import time
import bpy
import re



def ShowMessageBox(message = "", title = "Message Box", icon = 'INFO'):

	def draw(self, context):
		self.layout.label(text=message)

	bpy.context.window_manager.popup_menu(draw, title = title, icon = icon)




def find_collection(item):
	collections = item.users_collection
	if len(collections) > 0:
		return collections[0]
	return bpy.context.scene.collection


def make_collection(collection_name, parent_collection=False):
	if collection_name in bpy.data.collections: # Does the collection already exist?
		return bpy.data.collections[collection_name]
	else:
		new_collection = bpy.data.collections.new(collection_name)
		if parent_collection:
			parent_collection.children.link(new_collection) # Add the new collection under a parent
		return new_collection

def mooveToCollection(obj, target_Collection, unlink=False):
	obj_collection = find_collection(obj)

	target_Collection.objects.link(obj)  # put the cube in the new collection
	if unlink:
		obj_collection.objects.unlink(obj)  # remove it from the old collection




def distance(obj0, obj1):
	"get distance fonction"
	p0 = obj0.location
	p1 = obj1.location  

	return math.sqrt((p0[0] - p1[0])**2 + (p0[1] - p1[1])**2 + (p0[2] - p1[2])**2 )



def IKPLNodeTree():
	if 'IKPLCurveData' not in bpy.data.node_groups:
		ng = bpy.data.node_groups.new('IKPLCurveData', 'ShaderNodeTree')
		ng.use_fake_user = True
	return bpy.data.node_groups['IKPLCurveData'].nodes



def IKPLCurveData(curve_name):
	Disque = eval(bpy.context.scene.IKPLinks)
	if curve_name not in Disque:
		cn = IKPLNodeTree().new('ShaderNodeRGBCurve')
		cn.name = curve_name
	 


def rangef(step):
	if step == 0:
		return[0]
	liste=[]
	x = 0.0
	while x <= 1:
		x += step
		liste.append(x)
	return liste


def UpDateSplines(obj_data, plname, cList, curve_hand, mode, Radius, segment):  

	obj_data.splines.clear()

	def radi(liste):
		rad_curve = bpy.data.node_groups['IKPLCurveData'].nodes[plname].mapping
		rad_curve.initialize()

		rr = rangef(1/(len(liste) if len(liste) else 1))
		
		if (2, 82, 0) <= bpy.app.version:
			return np.array([rad_curve.evaluate(rad_curve.curves[3], x) for x in rr])
		else:
			return np.array([rad_curve.curves[3].evaluate(x) for x in rr])
	

	if mode == 0:#singel line
		Spline = obj_data.splines.new('BEZIER')
		Spline.resolution_u = segment
		Spline.bezier_points.add(len(cList[0])-1)

		if Radius:
			eval_radius = list(reversed(radi(cList[0]) * 4))

		for num in range( len(cList[0]) ):
			x, y, z = cList[0][num].location	
			Spline.bezier_points[num].co = (x, y, z) 
			if Radius:
				Spline.bezier_points[num].radius = eval_radius[num if num < len(cList[0]) else num - len(cList[0]) + num]

			Spline.bezier_points[num].handle_right_type = curve_hand 
			Spline.bezier_points[num].handle_left_type  = curve_hand 

			Spline.bezier_points[num].handle_right_type = curve_hand 
			Spline.bezier_points[num].handle_left_type  = curve_hand 

	if mode == 1:#Network
		sp_lines = []

		if Radius:
			eval_radius = list(reversed(radi(cList[0]) * 2))
		
		for pairs, step in zip(cList[0], cList[1]):
			Spline = obj_data.splines.new('BEZIER')
			Spline.resolution_u = segment
			Spline.bezier_points.add(1)

			for num in (0,1):
				x, y, z = pairs[num].location	
				Spline.bezier_points[num].co = (x, y, z) 

				if Radius: 
					if len(eval_radius) > step[num]:
						Spline.bezier_points[num].radius = eval_radius[step[num]]
				
				Spline.bezier_points[num].handle_right_type = 'VECTOR'  
				Spline.bezier_points[num].handle_left_type  = 'VECTOR' 

				Spline.bezier_points[num].handle_right_type = 'VECTOR'  
				Spline.bezier_points[num].handle_left_type  = 'VECTOR' 

	


def returnparent(obj):
	if obj.parent == None:
		return obj.name
	return returnparent(obj.parent )


def parentSorte(liste):
	C = bpy.context; D = bpy.data; lsort  = [];

	for name in liste:
		if name in C.scene.objects:
			hot = returnparent(C.scene.objects[name])
			lsort.append(hot) if not hot in lsort else None
	return lsort + [n for n in liste if not n in lsort]



def efacer(plname):
	"""del useless"""
	C = bpy.context
	D = bpy.data
	scene = C.scene
	Disque = parentSorte(eval(scene.IKPLinks))


	Disque.remove(plname) 
	scene.IKPLinks = str(Disque)

	ng   = bpy.data.node_groups #------------urv maping 
	ICD  = ng['IKPLCurveData'].nodes if 'IKPLCurveData' in ng else None
	ICD.remove(ICD[plname]) if ICD != None else None

	if getattr(scene,plname,False):
		Set = eval('bpy.context.scene.{p}[0]'.format( p = plname ) )

		if not Set.baking:
			bpy.ops.object.delete(use_global=False)

		exec( "del bpy.types.Scene.{p}".format( p = plname ) )
		print("del %s"%plname)



def Sort(data, scope, gama, units, target_ob, mode=0, dist=False, short=False):
	"""Flow sort from target object"""
	co_find = target_ob.location


	if not dist and not short and mode==0: return [data, []] #-------------------

	size = len(data)
	kd = mathutils.kdtree.KDTree(size)

	for i, v in enumerate(data):
		kd.insert(v.location, i)

	kd.balance()
	by_distance = []
	set_of_dist = []
	for i in kd.find_n(co_find, size):
		by_distance.append(data[i[1]])
		set_of_dist.append(data.index(data[i[1]]))

	rest = by_distance.copy()

	if dist and not short and mode==0: return [by_distance, []] #-------------------

	if mode==1: #-------------------------------- NETWORK
		out = [[],[]] # [list(paire), list(radius_values)]
		#set_of_dist = list(reversed(set_of_dist))

		pairs = [] 
		checkPairs = set()
		for i, v in enumerate(data):
			for p in kd.find_n(v.location, gama):
				if 0 < p[2] <= units and i!=p[1]:
						edge = sorted((i,p[1]))

						if not str(edge) in checkPairs:
							out[0].append([data[edge[0]], data[edge[1]]])
							out[1].append([set_of_dist.index(edge[0]), set_of_dist.index(edge[1])])
							checkPairs.add(str(edge)) 

		return out #--------------------------------


	viewed = [] 
	viewed_conection = []
	if short:
		while len(viewed)<len(data):
			"""Recurcively travel to nearests point"""
			if len(viewed): co_find = viewed[-1].location #start
			
			size = len(rest)
			kd = mathutils.kdtree.KDTree(size)
			
			for i, v in enumerate(rest):
				kd.insert(v.location, i)

			kd.balance()
			MAP = kd.find_n(co_find, size)

			idx = MAP[0][1] # [(co, index, dist),...]
			
			viewed_conection.append([rest[p[1]] for p in MAP[:gama] if p[2] <= units] )
			viewed.append(rest.pop(idx))				#gama 		 #unite
	

	if dist: 
		out=by_distance
	if short:
		out=viewed
	if not dist and not short: 
		out=data

	return [out , [viewed_conection]] #--------------------------------







def runikpl():
	C = bpy.context
	D = bpy.data
	scene = C.scene


	Disque = parentSorte(eval(scene.IKPLinks))

	for Set in Disque:

		if getattr(scene,  Set , None) == None :
			continue

		Set = eval('scene.{S}[0]'.format( S = Set ) )
		
		cible  = eval(Set.obj_target)	 #target object [[bo.name],[p sys]]
		plname = Set.ikpl_name			 #self link name

		
		if (not cible[0] in scene.objects) or (not plname in scene.objects) : efacer(plname);						continue;
		if (not cible[1] in list(( sp.name for sp in C.scene.objects[cible[0]].particle_systems))): efacer(plname);	continue;
		

		SYS = C.scene.objects[cible[0]].particle_systems[ cible[1] ] 
		if (SYS.point_cache.is_baking == True) or (SYS.settings.type != "EMITTER"):
			continue

		obj_link = scene.objects[plname]

		if obj_link.parent:
			if obj_link.parent.name in Disque:
				Set.ikpl_data = obj_link.parent.name
		else:
			Set.ikpl_data = "-DataStream-"

		mode   = Set.ikpl_mode		   #0 "line", 1 "Graph"
		scope  = Set.ikpl_scope		   #point visible par step
		units  = Set.ikpl_units		   #link distance
		handle = Set.curve_handle	   #curve handle
		dalink = Set.ikpl_data		   #data link
		Radius = Set.ikpl_CrvRadius	   #bool use radius
		gama   = Set.ikpl_branches	   #point colecté par step

		def load(cible):

			dp = bpy.context.evaluated_depsgraph_get()
			psys  = dp.objects[ cible[0] ].particle_systems[ cible[1] ].particles
			dtp   = dp.objects[ cible[0] ].particle_systems[ cible[1] ].settings.name
			alive = [p for p in psys if p.alive_state == 'ALIVE' ]
			
			return alive
		
							
		if Set.flow != None:
			target_ob = Set.flow
		else:
			dp = bpy.context.evaluated_depsgraph_get()
			target_ob = dp.objects[ cible[0] ]


		curve = Sort(list(reversed(load(cible))), scope, gama, units, target_ob, mode, Set.ikpl_distance, Set.ikpl_short	)
		UpDateSplines(obj_link.data, plname, curve, handle, mode, Radius, Set.segment)  
		bpy.context.view_layer.update()

		

